"""
Functions for plotting ants images
"""


__all__ = [
    "plot"
]

import fnmatch
import math
import os
import warnings

from matplotlib import gridspec
import matplotlib.pyplot as plt
import matplotlib.patheffects as path_effects
import matplotlib.lines as mlines
import matplotlib.patches as patches
import matplotlib.mlab as mlab
import matplotlib.animation as animation
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

import numpy as np
import ants
from ants.decorators import image_method

@image_method
def plot(
    image,
    overlay=None,
    blend=False,
    alpha=1,
    cmap="Greys_r",
    overlay_cmap="turbo",
    overlay_alpha=0.9,
    vminol=None,
    vmaxol=None,
    cbar=False,
    cbar_length=0.8,
    cbar_dx=0.0,
    cbar_vertical=True,
    axis=0,
    nslices=12,
    slices=None,
    ncol=None,
    slice_buffer=None,
    black_bg=True,
    bg_thresh_quant=0.01,
    bg_val_quant=0.99,
    domain_image_map=None,
    crop=False,
    scale=False,
    reverse=False,
    title=None,
    title_fontsize=20,
    title_dx=0.0,
    title_dy=0.0,
    filename=None,
    dpi=500,
    figsize=1.5,
    reorient=True,
    resample=True,
):
    """
    Plot an ANTsImage.

    Use mask_image and/or threshold_image to preprocess images to be be
    overlaid and display the overlays in a given range. See the wiki examples.

    By default, images will be reoriented to 'LAI' orientation before plotting.
    So, if axis == 0, the images will be ordered from the
    left side of the brain to the right side of the brain. If axis == 1,
    the images will be ordered from the anterior (front) of the brain to
    the posterior (back) of the brain. And if axis == 2, the images will
    be ordered from the inferior (bottom) of the brain to the superior (top)
    of the brain.

    ANTsR function: `plot.antsImage`

    Arguments
    ---------
    image : ANTsImage
        image to plot

    overlay : ANTsImage
        image to overlay on base image

    cmap : string
        colormap to use for base image. See matplotlib.

    overlay_cmap : string
        colormap to use for overlay images, if applicable. See matplotlib.

    overlay_alpha : float
        level of transparency for any overlays. Smaller value means
        the overlay is more transparent. See matplotlib.

    axis : integer
        which axis to plot along if image is 3D

    nslices : integer
        number of slices to plot if image is 3D

    slices : list or tuple of integers
        specific slice indices to plot if image is 3D.
        If given, this will override `nslices`.
        This can be absolute array indices (e.g. (80,100,120)), or
        this can be relative array indices (e.g. (0.4,0.5,0.6))

    ncol : integer
        Number of columns to have on the plot if image is 3D.

    slice_buffer : integer
        how many slices to buffer when finding the non-zero slices of
        a 3D images. So, if slice_buffer = 10, then the first slice
        in a 3D image will be the first non-zero slice index plus 10 more
        slices.

    black_bg : boolean
        if True, the background of the image(s) will be black.
        if False, the background of the image(s) will be determined by the
            values `bg_thresh_quant` and `bg_val_quant`.

    bg_thresh_quant : float
        if white_bg=True, the background will be determined by thresholding
        the image at the `bg_thresh` quantile value and setting the background
        intensity to the `bg_val` quantile value.
        This value should be in [0, 1] - somewhere around 0.01 is recommended.
            - equal to 1 will threshold the entire image
            - equal to 0 will threshold none of the image

    bg_val_quant : float
        if white_bg=True, the background will be determined by thresholding
        the image at the `bg_thresh` quantile value and setting the background
        intensity to the `bg_val` quantile value.
        This value should be in [0, 1]
            - equal to 1 is pure white
            - equal to 0 is pure black
            - somewhere in between is gray

    domain_image_map : ANTsImage
        this input ANTsImage or list of ANTsImage types contains a reference image
        `domain_image` and optional reference mapping named `domainMap`.
        If supplied, the image(s) to be plotted will be mapped to the domain
        image space before plotting - useful for non-standard image orientations.

    crop : boolean
        if true, the image(s) will be cropped to their bounding boxes, resulting
        in a potentially smaller image size.
        if false, the image(s) will not be cropped

    scale : boolean or 2-tuple
        if true, nothing will happen to intensities of image(s) and overlay(s)
        if false, dynamic range will be maximized when visualizing overlays
        if 2-tuple, the image will be dynamically scaled between these quantiles

    reverse : boolean
        if true, the order in which the slices are plotted will be reversed.
        This is useful if you want to plot from the front of the brain first
        to the back of the brain, or vice-versa

    title : string
        add a title to the plot

    filename : string
        if given, the resulting image will be saved to this file

    dpi : integer
        determines resolution of image if saved to file. Higher values
        result in higher resolution images, but at a cost of having a
        larger file size

    resample : bool
        if true, resample image if spacing is very unbalanced.

    Example
    -------
    >>> import ants
    >>> import numpy as np
    >>> img = ants.image_read(ants.get_data('r16'))
    >>> segs = img.kmeans_segmentation(k=3)['segmentation']
    >>> ants.plot(img, segs*(segs==1), crop=True)
    >>> ants.plot(img, segs*(segs==1), crop=False)
    >>> mni = ants.image_read(ants.get_data('mni'))
    >>> segs = mni.kmeans_segmentation(k=3)['segmentation']
    >>> ants.plot(mni, segs*(segs==1), crop=False)
    """
    if (axis == "x") or (axis == "saggittal"):
        axis = 0
    if (axis == "y") or (axis == "coronal"):
        axis = 1
    if (axis == "z") or (axis == "axial"):
        axis = 2

    def mirror_matrix(x):
        return x[::-1, :]

    def rotate270_matrix(x):
        return mirror_matrix(x.T)

    def rotate180_matrix(x):
        return x[::-1, ::-1]

    def rotate90_matrix(x):
        return x.T

    def reorient_slice(x, axis):
        if axis != 2:
            x = rotate90_matrix(x)
        if axis == 2:
            x = rotate270_matrix(x)
        x = mirror_matrix(x)
        return x


    # handle `image` argument
    if isinstance(image, str):
        image = ants.image_read(image)
    if not ants.is_image(image):
        raise ValueError("image argument must be an ANTsImage")

    if np.all(np.equal(image.numpy(), 0.0)):
        warnings.warn("Image must be non-zero. will not plot.")
        return

    # need this hack because of a weird NaN warning from matplotlib with overlays
    warnings.simplefilter("ignore")

    if (image.pixeltype not in {"float", "double"}) or (image.is_rgb):
        scale = False  # turn off scaling if image is discrete

    # handle `overlay` argument
    if overlay is not None:
        if isinstance(overlay, str):
            overlay = ants.image_read(overlay)
        if vminol is None:
            vminol = overlay.min()
        if vmaxol is None:
            vmaxol = overlay.max()
        if not ants.is_image(overlay):
            raise ValueError("overlay argument must be an ANTsImage")
        if overlay.components > 1:
            raise ValueError("overlay cannot have more than one voxel component")

        if not ants.image_physical_space_consistency(image, overlay):
            overlay = ants.resample_image_to_target(overlay, image, interp_type="nearestNeighbor")

        if blend:
            if alpha == 1:
                alpha = 0.5
            image = image * alpha + overlay * (1 - alpha)
            overlay = None
            alpha = 1.0

    # handle `domain_image_map` argument
    if domain_image_map is not None:
        tx = ants.new_ants_transform(
            precision="float",
            transform_type="AffineTransform",
            dimension=image.dimension,
        )
        image = ants.apply_ants_transform_to_image(tx, image, domain_image_map)
        if overlay is not None:
            overlay = ants.apply_ants_transform_to_image(
                tx, overlay, domain_image_map, interpolation="nearestNeighbor"
            )

    ## single-channel images ##
    if image.components == 1:

        # potentially crop image
        if crop:
            plotmask = image.get_mask(cleanup=0)
            if plotmask.max() == 0:
                plotmask += 1
            image = image.crop_image(plotmask)
            if overlay is not None:
                overlay = overlay.crop_image(plotmask)

        # potentially find dynamic range
        if scale == True:
            vmin, vmax = image.quantile((0.05, 0.95))
        elif isinstance(scale, (list, tuple)):
            if len(scale) != 2:
                raise ValueError(
                    "scale argument must be boolean or list/tuple with two values"
                )
            vmin, vmax = image.quantile(scale)
        else:
            vmin = None
            vmax = None

        # Plot 2D image
        if image.dimension == 2:

            img_arr = image.numpy()
            img_arr = rotate90_matrix(img_arr)

            if not black_bg:
                img_arr[img_arr < image.quantile(bg_thresh_quant)] = image.quantile(
                    bg_val_quant
                )

            if overlay is not None:
                ov_arr = overlay.numpy()
                mask = ov_arr == 0
                mask = np.ma.masked_where(mask == 0, mask)
                ov_arr = np.ma.masked_array(ov_arr, mask)
                ov_arr = rotate90_matrix(ov_arr)

            fig = plt.figure()
            if title is not None:
                fig.suptitle(
                    title, fontsize=title_fontsize, x=0.5 + title_dx, y=0.95 + title_dy
                )

            ax = plt.subplot(111)

            # plot main image
            im = ax.imshow(img_arr, cmap=cmap, alpha=alpha, vmin=vmin, vmax=vmax)

            if overlay is not None:
                im = ax.imshow(ov_arr, alpha=overlay_alpha, cmap=overlay_cmap,
                    vmin=vminol, vmax=vmaxol )

            if cbar:
                cbar_orient = "vertical" if cbar_vertical else "horizontal"
                fig.colorbar(im, orientation=cbar_orient)

            plt.axis("off")

        # Plot 3D image
        elif image.dimension == 3:
            # resample image if spacing is very unbalanced
            spacing = [s for i, s in enumerate(image.spacing) if i != axis]
            was_resampled = False
            if (max(spacing) / min(spacing)) > 3.0 and resample:
                was_resampled = True
                new_spacing = (1, 1, 1)
                image = image.resample_image(tuple(new_spacing))
                if overlay is not None:
                    overlay = overlay.resample_image(tuple(new_spacing))

            if reorient:
                image = image.reorient_image2("LAI")
            img_arr = image.numpy()
            # reorder dims so that chosen axis is first
            img_arr = np.rollaxis(img_arr, axis)

            if overlay is not None:
                if reorient:
                    overlay = overlay.reorient_image2("LAI")
                ov_arr = overlay.numpy()
                mask = ov_arr == 0
                mask = np.ma.masked_where(mask == 0, mask)
                ov_arr = np.ma.masked_array(ov_arr, mask)
                ov_arr = np.rollaxis(ov_arr, axis)

            if slices is None:
                if not isinstance(slice_buffer, (list, tuple)):
                    if slice_buffer is None:
                        slice_buffer = (
                            int(img_arr.shape[1] * 0.1),
                            int(img_arr.shape[2] * 0.1),
                        )
                    else:
                        slice_buffer = (slice_buffer, slice_buffer)
                nonzero = np.where(img_arr.sum(axis=(1, 2)) > 0.01)[0]
                min_idx = nonzero[0] + slice_buffer[0]
                max_idx = nonzero[-1] - slice_buffer[1]
                if min_idx > max_idx:
                    temp = min_idx
                    min_idx = max_idx
                    max_idx = temp
                if max_idx > nonzero.max():
                    max_idx = nonzero.max()
                if min_idx < 0:
                    min_idx = 0
                slice_idxs = np.linspace(min_idx, max_idx, nslices).astype("int")
                if reverse:
                    slice_idxs = np.array(list(reversed(slice_idxs)))
            else:
                if isinstance(slices, (int, float)):
                    slices = [slices]
                # if all slices are less than 1, infer that they are relative slices
                if sum([s > 1 for s in slices]) == 0:
                    slices = [int(s * img_arr.shape[0]) for s in slices]
                slice_idxs = slices
                nslices = len(slices)

            if was_resampled:
                # re-calculate slices to account for new image shape
                slice_idxs = np.unique(
                    np.array(
                        [
                            int(s * (image.shape[axis] / img_arr.shape[0]))
                            for s in slice_idxs
                        ]
                    )
                )

            # only have one row if nslices <= 6 and user didnt specify ncol
            if ncol is None:
                if nslices <= 6:
                    ncol = nslices
                else:
                    ncol = int(round(math.sqrt(nslices)))

            # calculate grid size
            nrow = math.ceil(nslices / ncol)
            xdim = img_arr.shape[2]
            ydim = img_arr.shape[1]

            dim_ratio = ydim / xdim
            fig = plt.figure(
                figsize=((ncol + 1) * figsize * dim_ratio, (nrow + 1) * figsize)
            )
            if title is not None:
                fig.suptitle(
                    title, fontsize=title_fontsize, x=0.5 + title_dx, y=0.95 + title_dy
                )

            gs = gridspec.GridSpec(
                nrow,
                ncol,
                wspace=0.0,
                hspace=0.0,
                top=1.0 - 0.5 / (nrow + 1),
                bottom=0.5 / (nrow + 1),
                left=0.5 / (ncol + 1),
                right=1 - 0.5 / (ncol + 1),
            )

            slice_idx_idx = 0
            for i in range(nrow):
                for j in range(ncol):
                    if slice_idx_idx < len(slice_idxs):
                        imslice = img_arr[slice_idxs[slice_idx_idx]]
                        imslice = reorient_slice(imslice, axis)
                        if not black_bg:
                            imslice[
                                imslice < image.quantile(bg_thresh_quant)
                            ] = image.quantile(bg_val_quant)
                    else:
                        imslice = np.zeros_like(img_arr[0])
                        imslice = reorient_slice(imslice, axis)

                    ax = plt.subplot(gs[i, j])
                    im = ax.imshow(imslice, cmap=cmap, vmin=vmin, vmax=vmax)

                    if overlay is not None:
                        if slice_idx_idx < len(slice_idxs):
                            ovslice = ov_arr[slice_idxs[slice_idx_idx]]
                            ovslice = reorient_slice(ovslice, axis)
                            im = ax.imshow(
                                ovslice, alpha=overlay_alpha, cmap=overlay_cmap,
                                    vmin=vminol, vmax=vmaxol )
                    ax.axis("off")
                    slice_idx_idx += 1

            if cbar:
                cbar_start = (1 - cbar_length) / 2
                if cbar_vertical:
                    cax = fig.add_axes([0.9 + cbar_dx, cbar_start, 0.03, cbar_length])
                    cbar_orient = "vertical"
                else:
                    cax = fig.add_axes([cbar_start, 0.08 + cbar_dx, cbar_length, 0.03])
                    cbar_orient = "horizontal"
                fig.colorbar(im, cax=cax, orientation=cbar_orient)

    ## multi-channel images ##
    elif image.has_components:
        raise Exception('Plotting images with components is not currently supported.')

    if filename is not None:
        filename = os.path.expanduser(filename)
        plt.savefig(filename, dpi=dpi, transparent=True, bbox_inches="tight")
        plt.close(fig)
    else:
        plt.show()

    # turn warnings back to default
    warnings.simplefilter("default")






__all__ = ['compose_displacement_fields']

import ants
from ants.internal import get_lib_fn


def compose_displacement_fields(displacement_field, 
                                warping_field):
    """
    Compose displacement fields.

    Arguments
    ---------
    displacement_field : ANTsImage displacement field
        displacement field

    warping_field : ANTsImage displacement field
        warping field


    Example
    -------
    >>> import ants
    """

    libfn = get_lib_fn('composeDisplacementFieldsD%i' % displacement_field.dimension)
    comp_field = libfn(displacement_field.pointer, warping_field.pointer)

    new_image = ants.from_pointer(comp_field).clone('float')
    return new_image





"""
ANTs ImageToImageMetric class
"""

import ants

class ANTsImageToImageMetric(object):
    """
    ANTsImageToImageMetric class
    """

    def __init__(self, metric):
        self._metric = metric
        self._is_initialized = False
        self.fixed_image = None
        self.fixed_mask = None
        self.moving_image = None
        self.moving_mask = None

    # ------------------------------------------
    # PROPERTIES
    @property
    def precision(self):
        return self._metric.precision

    @property
    def dimension(self):
        return self._metric.dimension

    @property
    def metrictype(self):
        return self._metric.metrictype.replace('ImageToImageMetricv4','')

    @property
    def is_vector(self):
        return self._metric.isVector == 1

    @property
    def pointer(self):
        return self._metric.pointer

    # ------------------------------------------
    # METHODS
    def set_fixed_image(self, image):
        """
        Set Fixed ANTsImage for metric
        """
        if not ants.is_image(image):
            raise ValueError('image must be ANTsImage type')

        if image.dimension != self.dimension:
            raise ValueError('image dim (%i) does not match metric dim (%i)' % (image.dimension, self.dimension))

        self._metric.setFixedImage(image.pointer, False)
        self.fixed_image = image

    def set_fixed_mask(self, image):
        """
        Set Fixed ANTsImage Mask for metric
        """
        if not ants.is_image(image):
            raise ValueError('image must be ANTsImage type')

        if image.dimension != self.dimension:
            raise ValueError('image dim (%i) does not match metric dim (%i)' % (image.dimension, self.dimension))

        self._metric.setFixedImage(image.pointer, True)
        self.fixed_mask = image

    def set_moving_image(self, image):
        """
        Set Moving ANTsImage for metric
        """
        if not ants.is_image(image):
            raise ValueError('image must be ANTsImage type')

        if image.dimension != self.dimension:
            raise ValueError('image dim (%i) does not match metric dim (%i)' % (image.dimension, self.dimension))

        self._metric.setMovingImage(image.pointer, False)
        self.moving_image = image

    def set_moving_mask(self, image):
        """
        Set Fixed ANTsImage Mask for metric
        """
        if not ants.is_image(image):
            raise ValueError('image must be ANTsImage type')

        if image.dimension != self.dimension:
            raise ValueError('image dim (%i) does not match metric dim (%i)' % (image.dimension, self.dimension))

        self._metric.setMovingImage(image.pointer, True)
        self.moving_mask = image

    def set_sampling(self, strategy='regular', percentage=1.):
        if (self.fixed_image is None) or (self.moving_image is None):
            raise ValueError('must set fixed_image and moving_image before setting sampling')

        self._metric.setSampling(strategy, percentage)

    def initialize(self):
        if (self.fixed_image is None) or (self.moving_image is None):
            raise ValueError('must set fixed_image and moving_image before initializing')

        self._metric.initialize()
        self._is_initialized = True

    def get_value(self):
        if not self._is_initialized:
            self.initialize()
        return self._metric.getValue()

    def __call__(self, fixed, moving, fixed_mask=None, moving_mask=None, sampling_strategy='regular', sampling_percentage=1.):
        self.set_fixed_image(fixed)
        self.set_moving_image(moving)

        if fixed_mask is not None:
            self.set_fixed_mask(fixed_mask)

        if moving_mask is not None:
            self.set_moving_mask(moving_mask)

        if (sampling_strategy is not None) or (sampling_percentage is not None):
            self.set_sampling(sampling_strategy, sampling_percentage)

        self.initialize()

        return self.get_value()

    def __repr__(self):
        s = "ANTsImageToImageMetric\n" +\
            '\t {:<10} : {}\n'.format('Dimension', self.dimension)+\
            '\t {:<10} : {}\n'.format('Precision', self.precision)+\
            '\t {:<10} : {}\n'.format('MetricType', self.metrictype)+\
            '\t {:<10} : {}\n'.format('IsVector', self.is_vector)
        return s





from .extract_image_patches import extract_image_patches
from .reconstruct_image_from_patches import reconstruct_image_from_patches

from .regression_match_image import regression_match_image

from .cropping_and_padding_utilities import crop_image_center
from .cropping_and_padding_utilities import pad_or_crop_image_to_size
from .cropping_and_padding_utilities import pad_image_by_factor

from .one_hot_segmentation import segmentation_to_one_hot
from .one_hot_segmentation import one_hot_to_segmentation

from .histogram_warp_image_intensities import histogram_warp_image_intensities
from .simulate_bias_field import simulate_bias_field
from .randomly_transform_image_data import randomly_transform_image_data
from .data_augmentation import data_augmentation



"""
Test ants_image.py

nptest.assert_allclose
self.assertEqual
self.assertTrue

"""

import os
import unittest
from common import run_tests

from tempfile import mktemp

import numpy as np
import numpy.testing as nptest

import ants


class TestClass_AntsImageIndexing(unittest.TestCase):

    def setUp(self):
        pass
    def tearDown(self):
        pass

    def test_pixeltype_2d(self):
        img = ants.image_read(ants.get_data('r16'))
        for ptype in ['unsigned char', 'unsigned int', 'float', 'double']:
            img = img.clone(ptype)
            self.assertEqual(img.pixeltype, ptype)
            img2 = img[:10,:10]
            self.assertEqual(img2.pixeltype, ptype)

    def test_pixeltype_3d(self):
        img = ants.image_read(ants.get_data('mni'))
        for ptype in ['unsigned char', 'unsigned int', 'float', 'double']:
            img = img.clone(ptype)
            self.assertEqual(img.pixeltype, ptype)
            img2 = img[:10,:10,:10]
            self.assertEqual(img2.pixeltype, ptype)
            img3 = img[:10,:10,10]
            self.assertEqual(img3.pixeltype, ptype)

    def test_2d(self):
        img = ants.image_read(ants.get_data('r16'))

        img2 = img[:10,:10]
        self.assertEqual(img2.dimension, 2)
        img2 = img[:5,:5]
        self.assertEqual(img2.dimension, 2)

        img2 = img[1:20,1:10]
        self.assertEqual(img2.dimension, 2)
        img2 = img[:5,:5]
        self.assertEqual(img2.dimension, 2)
        img2 = img[:5,4:5]
        self.assertEqual(img2.dimension, 2)

        img2 = img[5:5,5:5]

        # down to 1d
        arr = img[10,:]
        self.assertTrue(isinstance(arr, np.ndarray))

        # single value
        arr = img[10,10]

    def test_2d_image_index(self):
        img = ants.image_read(ants.get_data('r16'))
        idx = img > 200

        # acts like a mask
        img2 = img[idx]

    def test_3d(self):
        img = ants.image_read(ants.get_data('mni'))

        img2 = img[:10,:10,:10]
        self.assertEqual(img2.dimension, 3)
        img2 = img[:5,:5,:5]
        self.assertEqual(img2.dimension, 3)
        img2 = img[1:20,1:10,1:15]
        self.assertEqual(img2.dimension, 3)
        img2 = img[:5,:5,:5]
        self.assertEqual(img2.dimension, 3)

        # down to 2d
        img2 = img[10,:,:]
        self.assertEqual(img2.dimension, 2)
        img2 = img[:,10,:]
        self.assertEqual(img2.dimension, 2)
        img2 = img[:,:,10]
        self.assertEqual(img2.dimension, 2)
        img2 = img[10,:20,:20]
        self.assertEqual(img2.dimension, 2)
        img2 = img[:20,10,:]
        self.assertEqual(img2.dimension, 2)
        img2 = img[2:20,3:30,10]
        self.assertEqual(img2.dimension, 2)

        # down to 1d
        arr = img[10,:,10]
        self.assertTrue(isinstance(arr, np.ndarray))
        arr = img[10,:,5]
        self.assertTrue(isinstance(arr, np.ndarray))

        # single value
        arr = img[10,10,10]

    def test_double_indexing(self):
        img = ants.image_read(ants.get_data('mni'))
        img2 = img[20:,:,:]
        self.assertEqual(img2.shape, (162,218,182))

        img3 = img[0,:,:]
        self.assertEqual(img3.shape, (218,182))

    def test_reverse_error(self):
        img = ants.image_read(ants.get_data('mni'))
        with self.assertRaises(Exception):
            img2 = img[20:10,:,:]

    def test_2d_vector(self):
        img = ants.image_read(ants.get_data('r16'))
        img2 = img[:10,:10]

        img_v = ants.merge_channels([img])
        img_v2 = img_v[:10,:10]

        self.assertTrue(ants.allclose(img2, ants.split_channels(img_v2)[0]))

    def test_2d_vector_multi(self):
        img = ants.image_read(ants.get_data('r16'))
        img2 = img[:10,:10]

        img_v = ants.merge_channels([img,img,img])
        img_v2 = img_v[:10,:10]

        self.assertTrue(ants.allclose(img2, ants.split_channels(img_v2)[0]))
        self.assertTrue(ants.allclose(img2, ants.split_channels(img_v2)[1]))
        self.assertTrue(ants.allclose(img2, ants.split_channels(img_v2)[2]))

    def test_setting_3d(self):
        img = ants.image_read(ants.get_data('mni'))
        img2d = img[100,:,:]

        # setting a sub-image with an image
        img2 = img + 10
        img2[100,:,:] = img2d

        self.assertFalse(ants.allclose(img, img2))
        self.assertTrue(ants.allclose(img2d, img2[100,:,:]))

        # setting a sub-image with an array
        img2 = img + 10
        img2[100,:,:] = img2d.numpy()

        self.assertFalse(ants.allclose(img, img2))
        self.assertTrue(ants.allclose(img2d, img2[100,:,:]))

    def test_setting_2d(self):
        img = ants.image_read(ants.get_data('r16'))
        img2d = img[100,:]

        # setting a sub-image with an image
        img2 = img + 10
        img2[100,:] = img2d

        self.assertFalse(ants.allclose(img, img2))
        self.assertTrue(np.allclose(img2d, img2[100,:]))


    def test_setting_2d_sub_image(self):
        img = ants.image_read(ants.get_data('r16'))
        img2d = img[10:30,10:30]

        # setting a sub-image with an image
        img2 = img + 10
        img2[10:30,10:30] = img2d

        self.assertFalse(ants.allclose(img, img2))
        self.assertTrue(ants.allclose(img2d, img2[10:30,10:30]))

        # setting a sub-image with an array
        img2 = img + 10
        img2[10:30,10:30] = img2d.numpy()

        self.assertFalse(ants.allclose(img, img2))
        self.assertTrue(ants.allclose(img2d, img2[10:30,10:30]))

    def test_setting_correctness(self):

        img = ants.image_read(ants.get_data('r16')) * 0
        self.assertEqual(img.sum(), 0)

        img2 = img[10:30,10:30]
        img2 = img2 + 10
        self.assertEqual(img2.mean(), 10)

        img[:20,:20] = img2
        self.assertEqual(img.sum(), img2.sum())
        self.assertEqual(img.numpy()[:20,:20].sum(), img2.sum())
        self.assertNotEqual(img.numpy()[10:30,10:30].sum(), img2.sum())


    def test_slicing_3d(self):
        img = ants.image_read(ants.get_data('mni'))
        img2 = img[:10,:10,:10]
        img3 = img[10:20,10:20,10:20]

        self.assertTrue(ants.allclose(img2, img3))

        img_np = img.numpy()

        self.assertTrue(np.allclose(img2.numpy(), img_np[:10,:10,:10]))
        self.assertTrue(np.allclose(img[20].numpy(), img_np[20]))
        self.assertTrue(np.allclose(img[:,20:40].numpy(), img_np[:,20:40]))
        self.assertTrue(np.allclose(img[:,:,20:-2].numpy(), img_np[:,:,20:-2]))
        self.assertTrue(np.allclose(img[0:-1,].numpy(), img_np[0:-1,]))
        self.assertTrue(np.allclose(img[100,10:100,0:-1].numpy(), img_np[100,10:100,0:-1]))
        self.assertTrue(np.allclose(img[:,10:,30:].numpy(), img_np[:,10:,30:]))
        # if the slice returns 1D, it should be a numpy array already
        self.assertTrue(np.allclose(img[100:-1,30,40], img_np[100:-1,30,40]))

    def test_slicing_2d(self):
        img = ants.image_read(ants.get_data('r16'))

        img2 = img[:10,:10]

        img_np = img.numpy()

        self.assertTrue(np.allclose(img2.numpy(), img_np[:10,:10]))
        self.assertTrue(np.allclose(img[:,20:40].numpy(), img_np[:,20:40]))
        self.assertTrue(np.allclose(img[0:-1,].numpy(), img_np[0:-1,]))
        self.assertTrue(np.allclose(img[50:,10:-3].numpy(), img_np[50:,10:-3]))
        # if the slice returns 1D, it should be a numpy array already
        self.assertTrue(np.allclose(img[20], img_np[20]))
        self.assertTrue(np.allclose(img[100:-1,30], img_np[100:-1,30]))

if __name__ == '__main__':
    run_tests()


from .atropos import atropos
from .kmeans import kmeans_segmentation
from .functional_lung_segmentation import functional_lung_segmentation
from .fuzzy_spatial_cmeans_segmentation import fuzzy_spatial_cmeans_segmentation
from .kelly_kapowski import kelly_kapowski
from .joint_label_fusion import joint_label_fusion, local_joint_label_fusion
from .otsu import otsu_segmentation
from .prior_based_segmentation import prior_based_segmentation



__all__ = ["fit_bspline_displacement_field"]

import numpy as np

import ants
from ants.internal import get_lib_fn


def fit_bspline_displacement_field(displacement_field=None,
                                   displacement_weight_image=None,
                                   displacement_origins=None,
                                   displacements=None,
                                   displacement_weights=None,
                                   origin=None,
                                   spacing=None,
                                   size=None,
                                   direction=None,
                                   number_of_fitting_levels=4,
                                   mesh_size=1,
                                   spline_order=3,
                                   enforce_stationary_boundary=True,
                                   estimate_inverse=False,
                                   rasterize_points=False):

    """
    Fit a b-spline object to a dense displacement field image and/or a set of points
    with associated displacements and smooths them using B-splines.  The inverse
    can also be estimated..  This is basically a wrapper for the ITK filter

    https://itk.org/Doxygen/html/classitk_1_1DisplacementFieldToBSplineImageFilter.html}

    which, in turn is a wrapper for the ITK filter used for the function
    fit_bspline_object_to_scattered_data.

    ANTsR function: `fitBsplineToDisplacementField`

    Arguments
    ---------
    displacement_field : ANTs image
        Input displacement field.  Either this and/or the points must be specified.

    displacement_weight_image : ANTs image
        Input image defining weighting of the voxelwise displacements in the displacement_field.  I
        If None, defaults to identity weighting for each displacement.  Default = None.

    displacement_origins : 2-D numpy array
        Matrix (number_of_points x dimension) defining the origins of the input
        displacement points.  Default = None.

    displacements : 2-D numpy array
        Matrix (number_of_points x dimension) defining the displacements of the input
        displacement points.  Default = None.

    displacement_weights : 1-D numpy array
        Array defining the individual weighting of the corresponding scattered data value.
        Default = None meaning all values are weighted the same.

    origin : n-D tuple
        Defines the physical origin of the B-spline object.

    spacing : n-D tuple
        Defines the physical spacing of the B-spline object.

    size : n-D tuple
       Defines the size (length) of the B-spline object.  Note that the length of the
       B-spline object in dimension d is defined as
       spacing[d] * size[d]-1.

    direction : 2-D numpy array
       Booleans defining whether or not the corresponding parametric dimension is
       closed (e.g., closed loop).  Default = None.

    number_of_fitting_levels : integer
       Specifies the number of fitting levels.

    mesh_size : n-D tuple
       Defines the mesh size at the initial fitting level.

    spline_order : integer
       Spline order of the B-spline object.  Default = 3.

    enforce_stationary_boundary : boolean
       Ensure no displacements on the image boundary.  Default = True.

    estimate_inverse : boolean
       Estimate the inverse displacement field.  Default = False.

    rasterize_points : boolean
       Use nearest neighbor rasterization of points for estimating the
       field (potential speed-up).  Default = False.

    Returns
    -------
    Returns an ANTsImage.

    Example
    -------
    >>> import ants
    >>> import numpy as np
    >>> points = np.array([[-50, -50]])
    >>> deltas = np.array([[10, 10]])
    >>> bspline_field = ants.fit_bspline_displacement_field(
    >>>   displacement_origins=points, displacements=deltas,
    >>>   origin=[0.0, 0.0], spacing=[1.0, 1.0], size=[100, 100],
    >>>   direction=np.array([[-1, 0], [0, -1]]),
    >>>   number_of_fitting_levels=4, mesh_size=(1, 1))
    """

    if displacement_field is None and (displacement_origins is None or displacements is None):
        raise ValueError("Missing input.  Either a displacement field or input point set (origins + displacements) needs to be specified.")

    if displacement_field is None:
        if origin is None or spacing is None or size is None or direction is None:
            raise ValueError("If the displacement field is not specified, one must fully specify the input physical domain.")

    if displacement_field is not None and displacement_weight_image is None:
        displacement_weight_image = ants.make_image(displacement_field.shape, voxval=1,
            spacing=displacement_field.spacing, origin=displacement_field.origin,
            direction=displacement_field.direction, has_components=False, pixeltype='float')

    if displacement_field is not None:
        if origin is None:
            origin = displacement_field.origin
        if spacing is None:
            spacing = displacement_field.spacing
        if direction is None:
            direction = displacement_field.direction
        if size is None:
            size = displacement_field.shape

    dimensionality = None
    if displacement_field is not None:
        dimensionality = displacement_field.dimension
    else:
        dimensionality = displacement_origins.shape[1]
        if displacements.shape[1] != dimensionality:
            raise ValueError("Dimensionality between origins and displacements does not match.")

    if displacement_origins is not None:
        if displacement_weights is not None and (len(displacement_weights) != displacement_origins.shape[0]):
            raise ValueError("Length of displacement weights must match the number of displacement points.")
        else:
            displacement_weights = np.ones(displacement_origins.shape[0])

    if isinstance(mesh_size, int) == False and len(mesh_size) != dimensionality:
        raise ValueError("Incorrect specification for mesh_size.")

    if origin is not None and len(origin) != dimensionality:
        raise ValueError("Origin is not of length dimensionality.")

    if spacing is not None and len(spacing) != dimensionality:
        raise ValueError("Spacing is not of length dimensionality.")

    if size is not None and len(size) != dimensionality:
        raise ValueError("Size is not of length dimensionality.")

    if direction is not None and (direction.shape[0] != dimensionality and direction.shape[1] != dimensionality):
        raise ValueError("Direction is not of shape dimensionality x dimensionality.")

    # It would seem that pybind11 doesn't really play nicely when the
    # arguments are 'None'

    if origin is None:
        origin = np.empty(0)

    if spacing is None:
        spacing = np.empty(0)

    if size is None:
        size = np.empty(0)

    if direction is None:
        direction = np.empty((0, 0))

    if displacement_origins is None:
        displacement_origins = np.empty((0, 0))
        displacement_weights = np.empty(0)
    else:
        if displacement_weights is None:
            displacement_weights = np.repeat(1.0, displacement_origins.shape[0])

    number_of_control_points = list(np.array(mesh_size) + np.repeat(spline_order, dimensionality))

    bspline_field = None
    if displacement_field is not None:
        libfn = get_lib_fn("fitBsplineDisplacementFieldD%i" % (dimensionality))
        bspline_field = libfn(displacement_field.pointer, displacement_weight_image.pointer,
                              displacement_origins, displacements, displacement_weights,
                              origin, spacing, size, direction,
                              number_of_fitting_levels, number_of_control_points, spline_order,
                              enforce_stationary_boundary, estimate_inverse)
    elif displacement_field is None and displacements is not None:
        libfn = get_lib_fn("fitBsplineDisplacementFieldToScatteredDataD%i" % (dimensionality))
        bspline_field = libfn(displacement_origins, displacements, displacement_weights,
                              origin, spacing, size, direction,
                              number_of_fitting_levels, number_of_control_points, spline_order,
                              enforce_stationary_boundary, estimate_inverse, rasterize_points)


    bspline_displacement_field = ants.from_pointer(bspline_field).clone('float')
    return bspline_displacement_field

